package lsedit;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Rectangle;

import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;
import javax.swing.*;

// This is a layered pane so that edges draw on top of entities

public class Diagram extends /*JLayeredPane*/ JPanel implements Runnable, MouseMotionListener, DiagramCoordinates  
{
	class ExitFlag extends JComponent implements MouseListener
	{
		protected final static int EXIT_FLAG_DIM = 8;

		protected Diagram m_diagram;

		public ExitFlag(Diagram diagram)
		{
			super();
			m_diagram = diagram;
			addMouseListener(this);
		}

		public void setLocation(int x, int y)
		{
			setBounds(x, y, EXIT_FLAG_DIM, EXIT_FLAG_DIM);
		}

		public void paintComponent(Graphics g)
		{
//			System.out.println("Diagram.paintComponent()");

			super.paintComponent(g);

			/*  Draw a small mark as shown in top left of object

				x---
				|
				|   --------
				|  |        |
				   |        |
				   |        |
	  			   |  ----  |
				   |        |
				   |        |
				   |        |
				   |________|
			*/

			g.setColor(boxColour.darker());
			g.drawRect(0, 0, EXIT_FLAG_DIM, EXIT_FLAG_DIM);
			// g.setColor(Color.black);
			g.drawLine(2, 4, 6, 4);
//			System.out.println("Diagram.paintComponent() done");
		}

		// MouseListener interface

		public void mouseClicked(MouseEvent ev)	
		{
			m_diagram.navigateToDrawRootParent();
		}

		public void mouseEntered(MouseEvent e)
		{
			getLs().setCursor(Cursor.HAND_CURSOR);
		}

		public void mouseExited(MouseEvent e)
		{
			getLs().setCursor(Cursor.DEFAULT_CURSOR); 
		}

		public void mousePressed(MouseEvent ev)
		{
		}

		public void mouseReleased(MouseEvent ev)
		{
		}
	}

	// Final values

	public final static float BG = 0.75f;
	public final static String BG_STR = "0.75";
	public final static Color boxColour = Color.lightGray;

	protected final static int GAP = 5;
	protected final static int MARGIN = GAP*2;

	public final static String INSTANCE_ID				= "$INSTANCE";	
	public final static String INHERIT_RELN				= "$INHERIT";
	public final static String ALL_RELN					= "All";

	public final static String CONTAIN_ID				= "contain"; 
	public final static String EDGEMODE_ID				= "edgemode"; 
	public final static String TOPCLIENTS_ID			= "topclients"; 
	public final static String WANTCLIENTS_ID			= "wantclients"; 
	public final static String WANTSUPPLIERS_ID			= "wantsuppliers"; 
	public final static String WANTCARDINALS_ID			= "wantcardinals"; 
	public final static String SCALE_ID					= "scale";
	public final static String RELN_HIDDEN_ID			= "reln_hidden";
	public final static String NAVLINK_ID				= "navlink";

	public final static String RAW_HEADER				= "#TA_RAW#";

	public final static int UPDATE_FREQ = 250; 

	public final static int DEF_CAP = 2000; 
	public final static float DEF_LOAD = 0.5f;

	public final static int DEF_REL_CAP = 10000; 


	protected final static String ROOT_ID = "$ROOT";
	protected final static double MIN_REL_SCALE = 0.1;
	protected final static double MAX_REL_SCALE = 10.0;

	protected final static int MOVE_ZONE_MAX	= 20;
	protected final static String EDGE_SUFFIX = "$EDGESUFFIX$";

	// Values

	protected double	m_zoom_x = 1.0;		//	The zoom factor
	protected double	m_zoom_y = 1.0;

	protected Dimension	m_preferredSize = new Dimension();

	protected EntityInstance	m_rootInstance;				// The root instance of the graph
	protected EntityInstance	m_drawRoot;					// The current entity being drawn
	
	/* By putting the edges in a separate container we simplify the management of painting
		(1) We ensure that edges are painted after entities. This is because the m_edges
			components is added first.
		(2) We ensure that adding and deleting edges is efficient.
			This is because we can add at end.
	 */

	protected Container			m_edges     = null;

	/* Likewise for cardinals */

	protected Container			m_cardinals = null;

	protected int				m_drawRootLastWidth  = 0;
	protected int				m_drawRootLastHeight = 0;

	protected Hashtable m_entityClasses	  = new Hashtable(10);
	protected Hashtable m_relationClasses = new Hashtable(5);
	protected Hashtable m_entityInstances = new Hashtable(DEF_CAP, DEF_LOAD);
	protected Hashtable m_attrNames	      = new Hashtable(20);

	protected Vector	numToRel		  = new Vector(10);

	public    RelationClass		m_relationBaseClass; 
	public    EntityClass		m_entityBaseClass;

	protected int				numRelationClasses = 0;
	protected int				numVisibleRelationClasses = 0;

	protected EntityClass		defaultEntityClass = null;
	protected int				m_numEntityClasses = 0; 

	protected boolean			persistentQuery = false;
	protected boolean			universalScheme = false;
	protected boolean			allowRecursion  = true;
	protected boolean			readOnly        = false;
	protected boolean			m_loaded        = false;

	protected LandscapeViewerCore m_ls;

	protected Object			context;

	protected int				mode; 

	protected int				gridPixels = 1;

	protected int				m_preserve_entity_marks   = 0;
	protected int				m_preserve_relation_marks = 0;

	protected Attribute			relVisibilityAttr; 

	protected boolean			m_visibleEdges = true;
	protected boolean			m_drawEdges = true;
	protected boolean			m_drawBends = false;
	protected Color				xorMode	= null;

	protected int				newCnt = 0;					// Used to id new entities	
	protected Diagram			m_undoDg; 
	protected boolean			undoValid = false;

	protected int				timeStamp = 0;
	protected Vector			sourceList;
	protected Vector			sectionList;

	protected EntityInstance	m_keyEntity;				// The currently highlight entity if any else null

	protected boolean			changedFlag = false;

	protected ExitFlag			m_exitFlag;
	protected SupplierSet		m_supplierSet;
	protected ClientSet			m_clientSet;	

	protected Vector			m_suppliers;
	protected Vector			m_dClients;
	protected Vector			m_dSuppliers;

	protected boolean			m_suppliersSorted = false;

	protected boolean			allClients = true;
	protected boolean			allSuppliers = true;

	/* Must detect use of legacy coordinates and convert after diagram loaded
	 * since attributes may not be in tree order so won't know parent dimensions
	 * necessarily when doing conversion from legacy coordinates to relative ones
	 */

	boolean						m_uses_local_coordinates;

	/* Used in writing raw */

	private boolean				schemeSetup = false;
	private String				resString   = null;

	protected int				progressCount;

	// -----------------
	// Protected methods
	// -----------------

	protected void findClientsAndSuppliers() 
	{
		EntityInstance	drawRoot = m_drawRoot;
		boolean			showSuppliers;
		boolean			showClients;

		if (drawRoot == m_rootInstance) {
			showSuppliers = false;
			showClients   = false;
		} else {
			showSuppliers = m_ls.isShowSuppliers();
			showClients   = m_ls.isShowClients();
		}
		m_clientSet.removeAll();
		if (showClients) {
			m_clientSet.findClients(m_rootInstance, drawRoot);
		} 
		m_supplierSet.removeAll();
		if (showSuppliers) {
			m_supplierSet.findSuppliers(m_rootInstance, drawRoot);
		}
	}
		
	protected void updateProgress() 
	{
		progressCount++;

		StringBuffer sb = new StringBuffer("Loading: ");
		for (int i=0; i<progressCount; i++) {
			sb.append('*');
		}
		m_ls.doFeedback(new String(sb));
	}

	protected void setVisibilityFlags() 
	{
		Enumeration en;

		numVisibleRelationClasses = 0;

		for (en = enumRelations(); en.hasMoreElements(); ) {
			RelationClass rc = (RelationClass) en.nextElement();
			if (!excludeReln(rc)) {
				if (rc.isClassVisible()) {
					rc.setOrdinal(numVisibleRelationClasses++);
				}
			}
		}
		m_visibleEdges = (numVisibleRelationClasses > 0);
	}

	protected void registerAttrName(String id) 
	{
		if (m_attrNames.get(id) == null) {
			m_attrNames.put(id, id);
		}
	}

	protected void registerAttrNames(Attribute attr) {

		while(attr != null) {
			registerAttrName(attr.id);
			attr = attr.next;
		}

	}

	// After this method has been called however many times
	// need to fillLegendBox again

	protected RelationClass addRelationClass(String id) 
	{
		RelationClass rc = (RelationClass) m_relationClasses.get(id);		// Lookup in the hash table

		if (rc == null) {
			rc = new RelationClass(id, numRelationClasses, m_relationBaseClass, this);
			m_relationClasses.put(id, rc);
			numToRel.addElement(rc);
			numRelationClasses++;
		}
		return rc;
	}

	// After this method has been called however many times
	// need to fillLegendBox again

	protected EntityClass addEntityClass(String id) 
	{
		EntityClass ec = (EntityClass) m_entityClasses.get(id); 

		if (ec == null) {
			ec = new EntityClass(id, m_numEntityClasses++, m_entityBaseClass /* parent class */, this);
			m_entityClasses.put(id, ec);
			if (defaultEntityClass == null) {
				defaultEntityClass = ec;
			}
		}
		return ec;
	}

	protected EntityInstance addEntity(EntityClass ec, String eid) 
	{
		EntityInstance e = ec.newEntity(eid, this);
		e.setNid(m_entityInstances.size());
		m_entityInstances.put(eid, e);			// Add it to entity set
		// For now make it contained by root
		m_rootInstance.addContainment(e);
		return e;
	}

	protected void removeEntity(EntityInstance e)
	{
		e.getContainedBy().removeContainment(e);
		m_entityInstances.remove(e.getId());
	}



	protected boolean containsEntity(EntityInstance e) 
	{
		return m_entityInstances.contains(e.getId());
	}
	

	// Setup for universal scheme
	// This is employed if there are no SCHEME TUPLES

	protected void setupUniversalScheme() 
	{
		m_relationBaseClass.addRelation(m_entityBaseClass, m_entityBaseClass);
		universalScheme = true;
	}

	// Section processing methods

	protected void 
	processSchemeTuples(LandscapeTokenStream ts) throws IOException {

		for (;;) {

			Tuple tuple = ts.nextTuple();

			if (tuple == null) {
				return;
			}

//			System.out.println("Schema Tuples: " + tuple);

			RelationClass rc = null;

			boolean inheritance = tuple.token1.equals(INHERIT_RELN);			// $INHERIT

			if (inheritance) {
				if (tuple.token2.equals(EntityClass.ENTITY_BASE_CLASS_ID)) {	// $ENTITY
					ts.errorNS("Improper use of $ENTITY with $INHERIT");
					return;
				}
			} else { 
				rc = addRelationClass(tuple.token1);
			}

			EntityClass ec1 = addEntityClass(tuple.token2); 
			EntityClass ec2 = addEntityClass(tuple.token3);

			if (inheritance) {

				String msg = ec1.addParentClass(ec2);
				if (msg != null)
					ts.errorNS(msg);
			} else {
				rc.addRelation(ec1, ec2);		// If not already present
			}
		}
	}

	protected void processSchemeAttributes(LandscapeTokenStream ts) throws IOException 
	{

		for (;;) {

			AttributeRecord ar = ts.nextRecord();
			if (ar == null) {
				return;
			}

//			System.out.println("Schema Attributes: " + ar);

			if (ar.id.startsWith("(")) {
				StringTokenizer st = new StringTokenizer(ar.id, "() ");
				String rel = st.nextToken();

				RelationClass rc = (RelationClass) m_relationClasses.get(rel);

				if (rc != null) {
					Attribute attr = ar.attributes;
					while(attr != null) { 
						rc.addAttribute(attr);
						attr = attr.next;
					}
					registerAttrNames(ar.attributes);
				} else {
					ts.errorNS("Can't process record. Missing relation class '" + rel + "'");
				}
			} else {

				if (ar.id.equals(ROOT_ID)) {
					Attribute attr = ar.attributes;

					while(attr != null) {
						m_rootInstance.addAttribute(attr);
						attr = attr.next;
					}
					registerAttrNames(ar.attributes);
				} else { 

					EntityClass ec = (EntityClass) m_entityClasses.get(ar.id);

					if (ec != null) {
						Attribute attr = ar.attributes;
						while(attr != null) {
							ec.addAttribute(attr);
							attr = attr.next;
						}
						registerAttrNames(ar.attributes);
					} else {
						ts.errorNS("Can't process record. Missing entity class '" + ar.id + "'");
					}
				}
			}
		}
	}

	protected void processFactTuples(LandscapeTokenStream ts) throws IOException 
	{
		int ne = 0;
		int nr = 0;

		MsgOut.vprint("\nFACT TUPLE : ");

		for (;;) {
			Tuple tuple = ts.nextTuple();

			if (tuple == null) {
				return;
			}	

//			System.out.println("Fact Tuple: " + tuple);

			if (tuple.token1.equals(INSTANCE_ID)) {

				// 
				// $INSTANCE instanceId classId
				//

				ne++;
		
				if ((ne % UPDATE_FREQ) == 0) {
					MsgOut.vprint(".");
					updateProgress();
					m_ls.showInfo("Entities: " + ne);
				}		


				EntityInstance e = (EntityInstance) m_entityInstances.get(tuple.token2);

				if (e == null) {

					// New $INSTANCE

					EntityClass ec = (EntityClass) m_entityClasses.get(tuple.token3);


					if (ec != null) {

						e = ec.newEntity(tuple.token2, this);
						e.setNid(m_entityInstances.size());		// Add it to entity set

						m_entityInstances.put(tuple.token2, e);

						// For now make it contained by root

						m_rootInstance.addContainment(e);

					}  else {
						ts.errorNS("EntityClass '" + tuple.token3 + "' has not been declared");
					}
				} else {

					EntityClass ec = e.getEntityClass();

					if (ec.hasId(tuple.token3)) {
						ts.warning("Redeclaration of " + e.toString());
					} else {

						ts.errorNS("Attempt to declare " + e.getId() + " as instanceof " + tuple.token3 + ". Currently declared as instanceof " + ec.getId());
					}
				}
			} else {

				// Relation tuple definition
				//
				// relationClass entityInstance entityInstance
				//

				nr++;

				if ((nr % UPDATE_FREQ) == 0) {

					MsgOut.vprint(".");
					updateProgress();
					m_ls.showInfo("Relations: " + nr);
				}		



				RelationClass rc  = (RelationClass)	 m_relationClasses.get(tuple.token1);
				EntityInstance e1 = (EntityInstance) m_entityInstances.get(tuple.token2);
				EntityInstance e2 = (EntityInstance) m_entityInstances.get(tuple.token3);

				if (universalScheme) {

					// Implicit declaration upon use

					if (rc == null) {
						rc = m_relationBaseClass;
					}

					if (e1 == null) {
						e1 = m_entityBaseClass.newEntity(tuple.token2, this);
						e1.setNid(m_entityInstances.size());
						m_entityInstances.put(tuple.token2, e1); 
						m_rootInstance.addContainment(e1);
					}

					if (e2 == null) {
						e2 = m_entityBaseClass.newEntity(tuple.token3, this);
						e2.setNid(m_entityInstances.size());
						m_entityInstances.put(tuple.token3, e2);
						m_rootInstance.addContainment(e2);
					}
				}

				if (rc == null) {
					ts.errorNS("Can't process: " + "(" + tuple.token1 + " " + tuple.token2 + " " + tuple.token3 + ") - Missing '" + tuple.token1 + "'");
				} else if (e1 == null) {
					ts.errorNS("Can't process: " + "(" + tuple.token1 + " " + tuple.token2 + " " + tuple.token3 + ") - Missing '" + tuple.token2 + "'");
				} else if (e2 == null) {
					ts.errorNS("Can't process: " + "(" + tuple.token1 + " " + tuple.token2 + " " + tuple.token3 + ") - Missing '" + tuple.token3 + "'");
				} else if (!allowRecursion && e1 == e2) {
					ts.warning("Ignoring recursive relation: " + "(" + tuple.token1 + " " + tuple.token2 + " " + tuple.token3 + ")");
				} else {

					EntityClass ec1 = e1.getEntityClass();
					EntityClass ec2 = e2.getEntityClass();

					if (rc.isValidRelation(ec1, ec2)) {
						if (tuple.token1.equals(CONTAIN_ID)) {
							// Right now it is contained by root, so
							// remove from root's containment, and
							// make it contained by another e1.

							if (e2.isMarked(EntityInstance.IN_TREE_MARK)) {
								System.out.println("Entity " + e2 + " has multiple parents");
							} else {
								if (m_rootInstance.removeContainment(e2)) {
									e2.orMark(EntityInstance.IN_TREE_MARK);
									e1.addContainment(e2);
								} else {
									System.out.println("Entity " + e2 + " not in unassigned collection waiting containment");
							}	}
						} else {
							addEdge(e1, e2, rc);
						}
					} else {
						ts.errorNS("Invalid use of relation " + tuple.token1);
					}
				}
			}
		}
	}

	protected void processFactAttributes(LandscapeTokenStream ts) throws IOException 

	{
		int n = 0;

		MsgOut.vprint("\nFACT ATTRIBUTE : ");
		for (;;) {
			AttributeRecord ar = ts.nextRecord();
			if (ar == null) {
				break;
			}

//			System.out.println("Fact Attribute: " + ar);
				
			n++; 

			if ((n % UPDATE_FREQ) == 0) {
				MsgOut.vprint(".");
				updateProgress();
				m_ls.showInfo("Attr Records: " + n);
			}

			MsgOut.vprintln("Processing record: " + ar.id);

			if (ar.id.startsWith("(")) {
				// Attributes assigned to a relation tuple (rare)

				Tuple tuple = new Tuple(ar.id);

				RelationClass  rc  = (RelationClass)	m_relationClasses.get(tuple.token1);
				EntityInstance src = (EntityInstance)	m_entityInstances.get(tuple.token2);
				EntityInstance dst = (EntityInstance)	m_entityInstances.get(tuple.token3);

				if (rc == null) {
					ts.errorNS("Can't process record for " + ar.id + ". " + "Missing relation '" + tuple.token1 + "'");
				} else if (src == null) {
					ts.errorNS("Can't process record for " + ar.id + ". " + "Missing entity '" + tuple.token2 + "'");
				} else if (dst == null) {
					ts.errorNS("Can't process record for " + ar.id + ". " + "Missing entity '" + tuple.token3 + "'");
				} else {
					RelationInstance ri = src.getRelation(rc, dst);

					if (ri == null) {
						ts.errorNS("Can't process record. " + "Missing relation " + ar.id);
					} else {
						ri.assignAttributes(ar.attributes, m_entityInstances);
						registerAttrNames(ar.attributes);
					}
				}
			}  else {

				EntityInstance e;

				if (ar.id.equals(ROOT_ID)) {
					e = m_rootInstance;
				} else {
					e = (EntityInstance) m_entityInstances.get(ar.id);
				}
				if (e != null) {
					e.assignAttributes(ar.attributes);
					registerAttrNames(ar.attributes);
				} else if (universalScheme) {

					// Create it

					e = m_entityBaseClass.newEntity(ar.id, this);
					e.setNid(m_entityInstances.size());
					m_entityInstances.put(ar.id, e);

					m_rootInstance.addContainment(e);

					e.assignAttributes(ar.attributes);

					registerAttrNames(ar.attributes);

				} else {

					ts.errorNS("Can't process record. " + "Missing entity '" + ar.id + "'");
				}
			}
		}
	}



	// *** And now for output ***



	protected void writeRawPairs(DataOutputStream dos, Vector v) throws IOException 
	{
		Enumeration en;

		dos.writeInt(v.size());

		for (en = v.elements(); en.hasMoreElements(); ) {
			Integer[] pair = (Integer[]) en.nextElement();
			dos.writeInt(pair[0].intValue());
			dos.writeInt(pair[1].intValue());
		}
	}

	protected void writeRawTriples(DataOutputStream dos, Vector v) throws IOException
	{
		Enumeration en;

		dos.writeInt(v.size());

		for (en = v.elements(); en.hasMoreElements(); ) {
			Integer[] triple = (Integer[]) en.nextElement();
			dos.writeInt(triple[0].intValue());
			dos.writeInt(triple[1].intValue());
			dos.writeInt(triple[2].intValue());
		}
	}

	protected void writeRawAttributes(DataOutputStream dos, Vector v) throws IOException
	{
		dos.writeInt(v.size());

		Enumeration en = v.elements();
		while(en.hasMoreElements()) {
			Object[] obj = (Object []) en.nextElement();
			dos.writeInt(((Integer) obj[0]).intValue());
			Vector lv = (Vector) obj[1];
			dos.writeInt(lv.size());
			Enumeration en1 = lv.elements();
			while (en1.hasMoreElements()) {
				Object[] attr = (Object []) en1.nextElement();
				dos.writeInt(((Integer) attr[0]).intValue());
				int type = ((Integer) attr[1]).intValue();
				dos.writeInt(type);

				switch(type) 
				{
				case Attribute.NULL_TYPE:
					break;

				case Attribute.INT_TYPE:
					dos.writeInt(((Integer) attr[2]).intValue());
					break;

				case Attribute.DOUBLE_TYPE:
					dos.writeDouble(((Double) attr[2]).doubleValue());
					break;

				case Attribute.STRING_TYPE:
					dos.writeUTF((String) attr[2]);
					break;

				case Attribute.INT_LIST_TYPE:
					{
					int[] il = (int[]) attr[2];
					int num = il.length;

					dos.writeInt(num);

					for (int i=0; i<num; i++) {
						dos.writeInt(il[i]);
					}
					}
					break;

				case Attribute.DOUBLE_LIST_TYPE:
					{
					double[] dl = (double[]) attr[2];
					int num = dl.length;

					dos.writeInt(num);
					for (int i=0; i<num; i++) {
						dos.writeDouble(dl[i]);
					}
					}
					break; 

				case Attribute.STRING_LIST_TYPE:
					{
					String[] sl = (String[]) attr[2];
					int num = sl.length;

					dos.writeInt(num);
					for (int i=0; i<num; i++) {
						dos.writeUTF(sl[i]);
					}
					}
					break;
				}
			}
		}
	}

	protected void writeSchemeTuples(PrintStream ps) throws IOException {
		ps.print("// Landscape TA file\n\n");
		ps.print("SCHEME TUPLE :\n\n// The ERD\n\n"); 

		Enumeration enum1 = enumEntityClasses();

		while(enum1.hasMoreElements()) {
			EntityClass ec = (EntityClass) enum1.nextElement();
			if (ec != m_entityBaseClass) {
				Enumeration en = ec.getParentElements();
				while (en.hasMoreElements()) {
					EntityClass parent = (EntityClass) en.nextElement();
					if (parent != m_entityBaseClass) {
						ps.print(INHERIT_RELN + " " + ec.getId() + " " + parent.getId() + "\n");
					}
				}
			}
		}
		ps.print("\n");
		enum1 = enumRelations(); // Output in same order as input
		while(enum1.hasMoreElements()) {
			RelationClass rc = (RelationClass) enum1.nextElement();
			rc.writeRelations(ps); 
		}
	}

	/* Write out:
	 *
     * #1#
	 * cnt [id/parentid]*
     * cnt [id/from/to]*
     *
     */

	protected void writeSchemeTuplesRaw(DataOutputStream dos, Hashtable stringTable) throws IOException
	{
		Enumeration en, en1;

		dos.writeUTF("#1#");

		Vector v = new Vector();

		for (en = enumEntityClasses(); en.hasMoreElements(); ) {
			EntityClass ec = (EntityClass) en.nextElement();
			if (ec != m_entityBaseClass) {
				for (en1 = ec.getParentElements(); en1.hasMoreElements(); ) {
					EntityClass parent = (EntityClass) en1.nextElement();
					if (parent != m_entityBaseClass) {
						Integer[] pair = new Integer[2];
						pair[0] = (Integer) stringTable.get(ec.getId());
						pair[1] = (Integer) stringTable.get(parent.getId());
						v.addElement(pair);
					}
				}
			}
		}
		writeRawPairs(dos, v);

		v = new Vector();

		for (en = enumRelations(); en.hasMoreElements(); ) {	// Output in same order as input
			RelationClass rc = (RelationClass) en.nextElement();
			rc.getRelationsRaw(v, stringTable);
		}
		writeRawTriples(dos, v);
	}

	protected void processRawSchemeTuples(DataInputStream dis, Vector st) throws IOException 
	{
		// Read inherits

		int num = dis.readInt();
		for (int i=0; i<num; i++) {
			String child = (String) st.elementAt(dis.readInt());
			String parent = (String) st.elementAt(dis.readInt());
			EntityClass ec1 = addEntityClass(child);
			EntityClass ec2 = addEntityClass(parent);
			ec1.addParentClass(ec2);
		}

		num = dis.readInt();
		for (int i=0; i<num; i++) {
			String rcid	 = (String) st.elementAt(dis.readInt());
			String ecid1 = (String) st.elementAt(dis.readInt());
			String ecid2 = (String) st.elementAt(dis.readInt());

			RelationClass rc = addRelationClass(rcid);
			EntityClass ec1	 = addEntityClass(ecid1);
			EntityClass ec2	 = addEntityClass(ecid2);

			rc.addRelation(ec1, ec2);
		}
	}

	protected void writeSchemeAttributes(PrintStream ps) throws IOException 
	{
		Enumeration en;

		ps.print("\n\nSCHEME ATTRIBUTE :\n\n");
		ps.print("// EntityClass attributes\n\n");
		// Write attributes for entity classes 

		for (en = m_entityClasses.elements(); en.hasMoreElements(); ) {
			EntityClass ec = (EntityClass) en.nextElement(); 
			ec.writeAttributes(ps);
		}

		// Write attributes for relation classes 

		for (en = m_relationClasses.elements(); en.hasMoreElements(); ) {
			RelationClass rc = (RelationClass) en.nextElement();
			rc.writeAttributes(ps);
		}
	}

	protected void writeSchemeAttributesRaw(DataOutputStream dos, Hashtable stringTable) throws IOException
	{
		Enumeration en;

		// Write attributes for entity classes 
		dos.writeUTF("#2#");
		Vector v = new Vector();
		for (en = m_entityClasses.elements(); en.hasMoreElements(); ) {
			EntityClass ec = (EntityClass) en.nextElement(); 
			ec.getAttributesRaw(v, stringTable);
		}
		writeRawAttributes(dos, v);

		// Write attributes for relation classes 

		v = new Vector();

		for (en = m_relationClasses.elements(); en.hasMoreElements(); ) {
			RelationClass rc = (RelationClass) en.nextElement();
			rc.getAttributesRaw(v, stringTable);
		}
		writeRawAttributes(dos, v);
	}

	protected void processRawSchemeAttributes(DataInputStream dis, Vector st) throws IOException 
	{
		int num = dis.readInt();

		for (int i=0; i<num; i++) {
			String ecid = (String) st.elementAt(dis.readInt());
			EntityClass ec = (EntityClass) m_entityClasses.get(ecid);
			processRawAttributes(ec, dis, st);
		}

		num = dis.readInt();

		for (int i=0; i<num; i++) {
			String rcid = (String) st.elementAt(dis.readInt());
			RelationClass rc = (RelationClass) m_relationClasses.get(rcid);
			processRawAttributes(rc, dis, st);
		}
	}

	protected void writeFactAttributes(PrintStream ps) throws IOException
	{
		ps.print("\n\nFACT ATTRIBUTE :\n\n"); 
		m_rootInstance.writeAttributes(ps); 
	}

	protected void writeFactTuples(PrintStream ps) throws IOException 
	{
		ps.print("\n\nFACT TUPLE :\n\n"); 
		ps.print("// Instances of entity classes\n\n");
		m_rootInstance.writeInstances(ps);
		m_rootInstance.writeRelations(ps); 
	}



	protected void
	writeFactTuplesRaw(DataOutputStream dos, Hashtable stringTable)	throws IOException
	{
		dos.writeUTF("#3#");
		Vector v = new Vector();
		m_rootInstance.getInstancesRaw(v, stringTable);
		writeRawPairs(dos, v);
		v = new Vector();
		m_rootInstance.getRelationsRaw(v, stringTable, (Integer) stringTable.get(CONTAIN_ID)); 
		writeRawTriples(dos, v);
	}

	protected void writeFactAttributesRaw(DataOutputStream dos, Hashtable stringTable)	throws IOException
	{
		dos.writeUTF("#4#");
		Vector v = new Vector();
		m_rootInstance.getAttributesRaw(v, stringTable);
		writeRawAttributes(dos, v);
	}





	protected Vector readStringTable(DataInputStream dis) throws IOException 
	{
		Vector st = new Vector();
		int num = dis.readInt();

		for (int i=0; i<num; i++) {
			st.addElement(dis.readUTF());
		}

/*
		Enumeration en = st.elements();
		while(en.hasMoreElements()) {
			String str = (String) en.nextElement();
			System.out.println(str);
		}
*/
		return st;
	}

	protected void processRawAttributes(LandscapeObject lo, DataInputStream dis, Vector st) throws IOException
	{
		int na = dis.readInt();

		for (int i=0; i<na; i++) {
			String aid = (String) st.elementAt(dis.readInt());
			int type = dis.readInt();

			registerAttrName(aid);

			boolean rc;

			switch(type)
			{
			case Attribute.NULL_TYPE:
				rc = lo.addRawAttribute(aid);
				break;

			case Attribute.INT_TYPE:
				rc = lo.addRawAttribute(aid, dis.readInt());
				break;

			case Attribute.DOUBLE_TYPE:
				rc = lo.addRawAttribute(aid, dis.readDouble());
				break;
				
			case Attribute.STRING_TYPE:
				rc = lo.addRawAttribute(aid, dis.readUTF());
				break;

			case Attribute.INT_LIST_TYPE:
			{
				int num = dis.readInt();

				int[] il = new int[num];

				for (int j=0; j<num; j++) {
					il[j] = dis.readInt();
				}
				rc = lo.addRawAttribute(aid, il);
				break;
			}
			case Attribute.DOUBLE_LIST_TYPE:
			{
				int num = dis.readInt();
				double[] dl = new double[num];
				for (int j=0; j<num; j++) {
					dl[j] = dis.readDouble();
				}
				rc = lo.addRawAttribute(aid, dl);
				break;
			}
			case Attribute.STRING_LIST_TYPE:
			{
				int num = dis.readInt();
				String[] sl = new String[num];

				for (int j=0; j<num; j++) {
					sl[j] = dis.readUTF();
				}
				rc = lo.addRawAttribute(aid, sl);
				break;
			}
			default:
				rc = false;
				MsgOut.println("Unknown attribute type: " + type);
				break;
			}

			if (!rc) {
				MsgOut.println("Unhandled attribute: " + aid + " - type: " + type);
			}
		}
	}

	protected void processRawFactTuples(DataInputStream dis, Vector st) throws IOException 
	{
		// Read instances

		int num = dis.readInt();
		EntityClass ec	 = null;
		int			lind = -1;

		for (int i=0; i<num; i++) {
			String eid = (String) st.elementAt(dis.readInt());
			int ind = dis.readInt();
			if (ind != lind) {
				String ecid = (String) st.elementAt(ind);
				ec = (EntityClass) m_entityClasses.get(ecid);
				lind = ind;
			}
			addEntity(ec, eid);

			if ((i % UPDATE_FREQ) == 0) {
				MsgOut.vprint(".");
				updateProgress();
				m_ls.showInfo("Entities: " + i);
			}	
		}

		// Read edge tuples

		num = dis.readInt();
		RelationClass crc = (RelationClass) m_relationClasses.get(CONTAIN_ID);
		RelationClass	rc	   = null;
		int				lrcind = -1;
		EntityInstance	e1	   = null;
		int				leind  = -1;	

		for (int i=0; i<num; i++) {
			int rcind = dis.readInt();
			int eind  = dis.readInt();

			if (rcind != lrcind) {
				String rcid = (String) st.elementAt(rcind);
				rc = (RelationClass) m_relationClasses.get(rcid);
				lrcind = rcind;
			}
			if (eind != leind) {		
				String eid1 = (String) st.elementAt(eind);

				e1 = (EntityInstance) m_entityInstances.get(eid1);
				leind = eind;
			}

			String eid2 = (String) st.elementAt(dis.readInt());
			EntityInstance e2 = (EntityInstance) m_entityInstances.get(eid2);

			if (rc == crc) {
				// Right now it is contained by root, so
				// remove from root's containment, and
				// make it contained by e1.

				if (m_rootInstance.removeContainment(e2)) {
					e1.addContainment(e2);
				}
			} else {
				addEdge(e1, e2, rc);
			}
			if ((i % UPDATE_FREQ) == 0) {
				MsgOut.vprint(".");
				updateProgress();
				m_ls.showInfo("Relations: " + i);
			}
		}
	}

	protected void processRawFactAttributes(DataInputStream dis, Vector st) throws IOException 
	{
		int num = dis.readInt();

		for (int i=0; i<num; i++) {
			String eid = (String) st.elementAt(dis.readInt());
			EntityInstance e = (EntityInstance) m_entityInstances.get(eid);
			processRawAttributes(e, dis, st);
			if ((i % UPDATE_FREQ) == 0) {
				MsgOut.vprint(".");
				updateProgress();
				m_ls.showInfo("Attr records: " + i);
			}
		}
	}

	protected void parseRawStream(DataInputStream dis, String src, URL context)
	{
		try {
			String hdr = dis.readUTF();
			if (hdr.equals(RAW_HEADER)) {
				Vector stringTable = readStringTable(dis);
				for (;;) {
					String sec = dis.readUTF();
					if (sec.equals("#END#"))
						break;
					if (sec.equals("#1#")) 
						processRawSchemeTuples(dis, stringTable);
					else if (sec.equals("#2#"))
						processRawSchemeAttributes(dis, stringTable);
					else if (sec.equals("#3#"))
						processRawFactTuples(dis, stringTable);
					else if (sec.equals("#4#"))
						processRawFactAttributes(dis, stringTable);
					else  
						MsgOut.println(src + "- Bad section: " + sec);
				}
			} else {
				resString = "Not a raw TA file";
			}
		}
		catch (Exception e) {
			System.out.println(e);
			resString = e.toString();
		}
	}

	protected boolean isRaw(BufferedInputStream bis) throws IOException 
	{
		bis.mark(100);

		bis.read();
		bis.read();

		for (int i = 0; i<RAW_HEADER.length(); i++) {
			int c = bis.read();
			if (c != RAW_HEADER.charAt(i)) {
				bis.reset();
				return false;
			}
		}
		bis.reset();
		return true;
	}

	protected void computeRelCoordinates()
	{
		if (m_rootInstance != null) {
			m_rootInstance.computeRelCoordinates(m_rootInstance.xRelLocal(), m_rootInstance.yRelLocal(), m_rootInstance.widthRelLocal(), m_rootInstance.heightRelLocal() );
	}	}

	protected void parseStream(InputStream is, String src, URL context) {

		MsgOut.vprintln("Parse TA file: " + src);

		m_uses_local_coordinates = false;

		try {
			BufferedInputStream bis = new BufferedInputStream(is);

			if (isRaw(bis)) {
				DataInputStream dis = new DataInputStream(bis);
				parseRawStream(dis, src, context);
				dis.close();
				bis.close();
				is.close();
				return;
			}

			LandscapeTokenStream ts = new LandscapeTokenStream(bis, src);

			for (;;) {
				int sec = ts.nextSection();

				if (sec == LandscapeTokenStream.EOF) {
					MsgOut.vprintln("");
					break;
				}
				try { 
					switch(sec) {

					case LandscapeTokenStream.SCHEME_TUPLE:

						// regSrcInfo(src, sec);

						schemeSetup = true;
						universalScheme = false;

						processSchemeTuples(ts);
						updateProgress();
						break;



					case LandscapeTokenStream.SCHEME_ATTRIBUTE:

						// regSrcInfo(src, sec);

						if (!schemeSetup) {
							schemeSetup = true;
							setupUniversalScheme();
						}
						processSchemeAttributes(ts);
						updateProgress();
						break;



					case LandscapeTokenStream.FACT_TUPLE:

						// regSrcInfo(src, sec);

						if (!schemeSetup) {
							schemeSetup = true;
							setupUniversalScheme();
						}

						processFactTuples(ts);
						break;



					case LandscapeTokenStream.FACT_ATTRIBUTE:

						// int srcIndex = regSrcInfo(src, sec);

						processFactAttributes(ts);
						break;


					case LandscapeTokenStream.INCLUDE_FILE:

						if (context == null) {
							parseFile(ts.getIncludeFile(), null);
						} else {
							parseURL(ts.getIncludeFile(), context);
						}
						MsgOut.vprintln("Back to TA file: " + src);
						break; 
					}

				}

				catch (IOException e) {

					MsgOut.println("IO error reading landscape"); 
					resString = e.toString();
					break; 
				}
			}

			bis.close(); 
			is.close();

			if (m_uses_local_coordinates) {
				computeRelCoordinates();
			}
		}

		catch (Exception e) {
			resString = e.toString();
			System.out.println("Parse error: " + resString);
		}


	} // end parseStream() 



	protected URL parseURL(String taURL, URL context) {

		progressCount = 0;

		updateProgress();

		try {

			URL lsURL;

			if (context == null) {
				lsURL = new URL(taURL);

			} else {
				lsURL = new URL(context, taURL);
			}

			MsgOut.dprintln("Opening URL: " + taURL);
			InputStream is = lsURL.openStream();
			MsgOut.dprintln("opened");
			parseStream(is, taURL, lsURL);
			return lsURL; 
		}
		catch(Exception e) {
			resString = e.toString();
			return null;
		}
	}


	// Called from diagram.loadDiagram to parse a file or an include within a file

	protected File parseFile(String taFile, File context) {

		progressCount = 0;

		updateProgress();

		try {
			File file;

			if (context == null)
				file = new File(taFile);
			else {
				String dir = getDir(context);

				file = new File(dir, taFile);
			}

			FileInputStream fis = new FileInputStream(file);

			parseStream(fis, taFile, null);
			return file;

		} catch (Exception e) {
			resString = e.toString();
			return null;
		}
	}

	protected void processInternalBuffer(String taFile) 
	{
		InternalBufferStream iis = m_ls.getInternalBufferStream();

		if (iis == null) {
			resString = "Internal landscape not available";
			return;
		}

		parseRawStream(iis, taFile, null);
		readOnly = true;
	}

	// Non-first order attributes of $ROOT

	protected final static String g_rootAttributes[] =
		{ SCALE_ID, 
		  RELN_HIDDEN_ID, 
		  EDGEMODE_ID, 
		  NAVLINK_ID, 
		  TOPCLIENTS_ID,
		  WANTCLIENTS_ID,
		  WANTSUPPLIERS_ID,
		  WANTCARDINALS_ID
		};

	public void prepostorder()
	{
		m_rootInstance.prepostorder(1);
	}

	private void setRootInstance(EntityInstance rootInstance)
	{
		m_rootInstance = rootInstance;
		prepostorder();
	}

	public void setLoaded(boolean value)
	{
		m_loaded = value;
	}

/* 
	What a cautionary tale this is for a software engineering course..

	1. Somewhere in Java AWT's history someone had the bright idea of changing
	   the AWT framework so that every component could have a mouse cursor
	   associated with it instead of just the frame managing what the current
	   mouse cursor object should be at any given time.  The idea was that as
	   you moved the mouse over different objects each object would have the
	   capability of explicitly indicating what the cursor associated with the
	   mouse should be, while it hovered over that object.

	2. This means that when ever a new visible object is added to a container,
	   AWT must handle the possibility that the cursor happens to be over this
	   new object, in which case this new object should if appropriate change
	   the appearance of that mouse cursor immediately.

	3. So whenever an object is added to the AWT containment hierarchy, there
	   is an invocation to that object to change the system cursor immediately.

	   [This by the way is a less than smart implementation.. The smart thing
		to tell the component is change the cursor if the mouse pointer is over
		it and it actually cared what the cursor  should be in that unusual case]

	4. The object handles this request by walking all the way up the containment
	   heirarchy until it hits a component which actually cares what the cursor
	   looks like (in lsedit's case the topmost frame object, since tha ability
	   to associate cursors directly with objects is not used).

	5. This cursor isn't changed but the frame object doesn't know that. That
	   is because any component can over ride the frame's opinion as to what
	   the cursor should be (not that any does in lsedit).

	6. So the frame object starts off by asking itself what the mouse pointer
	   is over. To do this it walks back down through the containment heirarchy
	   asking each level if it contains the current point the mouse pointer is
	   over, recursively.  This comes close to search all components, if the
	   mouse pointer happens to be over the diagram, since entities and edges
	   dominate the number of components.

	7. Now containment is not a trivial exercise for edges.  At a minimum we
	   need to decide if the given point is within 'n' pixels of any part of
	   the line, the arrow head, etc.

	8. Having identified the deepest thing the mouse pointer is over, we walk up
	   finding the nearest component above which wishes to control the cursor
	   image, arriving back at the frame.  The frame then dutifully sets the
	   mouse cursor to what it already was.


	Now this is repeated every time we add a component.  So it is clearly an
	O(N^2) operation for N components.  So when displaying a large number of
	edges (order the number in linux or whatever) the cost of JAVA painting
	the cursor image associated with the mouse pointer, becomes prohibitive
	as the actual graph image is constructed line by line.

	I worked this out by narrowing down the performance problems when displaying
	a large ta file to: White cursor mouse arrow left on screen until lsedit
	draws = 40 seconds, while time taken was a mere 4 seconds if I cleverly fired
	up lsedit in a manner which left the cursor off the lsedit screen, somewhere
	down at the bottom of my actual screen.

	Needless to say I at first found this behaviour quite astonishing.

	Work around to correct this <<bug>> was to add to the diagram.java class
	the following special rule about what the diagram contains when not
	loaded -- nothing.

	Ian Davis
*/

	public boolean contains(int x, int y)
	{
		return (m_loaded && super.contains(x, y));
	}

	public void setPreserveEntityMarks(int value)
	{
		m_preserve_entity_marks = value;
	}

	public void setPreserveRelationMarks(int value)
	{
		m_preserve_relation_marks = value;
	}

	// For debugging
	
	
    protected void paintChildren(Graphics g) 
    {
    	super.paintChildren(g);
    }

	// For debugging
	    	
    public void repaint() 
    {
    	super.repaint();
    }
    
	// --------------
	// Public methods
	// --------------

	// This constructor is only called to create copies of diagrams
	// They are never active so don't need to assign them listeners

	public Diagram() 
	{
		setLayout(null);
		setLocation(0,0);
		setDoubleBuffered(true);
	}

	public Component add(Component c)
	{
		if (c instanceof RelationInstance) {
			return(m_edges.add(c));
		}
		if (c instanceof Cardinal) {
			return(m_cardinals.add(c));
		}
		return(super.add(c));
	}

	public void remove(Component c)
	{
		if (c instanceof RelationInstance) {
			m_edges.remove(c);
			return;
		}
		if (c instanceof Cardinal) {
			m_cardinals.remove(c);
			return;
		}
		super.remove(c);
	}

	public Diagram(LandscapeViewerCore ls, boolean makeBackup) 
	{
		m_ls = ls;

		setLayout(null);
		setLocation(0, 0);

		// Create Diagram for undo storage

		m_undoDg = new Diagram();

		m_exitFlag = new ExitFlag(this);
		m_exitFlag.setLocation(3, 3);

		// Create root entity

		setRootInstance(new BaseEntity(null, ROOT_ID, this));																				// $ROOT

		m_entityInstances.put(ROOT_ID, m_rootInstance);

		for (int i = 0; i < g_rootAttributes.length; i++) {
			m_rootInstance.addAttribute( new Attribute(g_rootAttributes[i], null));
		}

		AttributeValueItem av = new AttributeValueItem(BG_STR);																			// 0.75

		av.next = new AttributeValueItem(BG_STR);
		av.next.next = new AttributeValueItem(BG_STR);
		m_rootInstance.addAttribute(new Attribute("color", av));																			// $ROOT (0.75 0.75 0.75)

		// Need a navlink attribute on rootInstance

		m_entityBaseClass = new EntityClass(EntityClass.ENTITY_BASE_CLASS_ID, m_numEntityClasses++, null, this);							// $ENTITY
		m_entityClasses.put(EntityClass.ENTITY_BASE_CLASS_ID, m_entityBaseClass);

		m_relationBaseClass =	 new RelationClass(RelationClass.RELATION_BASE_CLASS_ID, numRelationClasses++, null /* parent */, this);	// $RELATION
		m_relationClasses.put(RelationClass.RELATION_BASE_CLASS_ID, m_relationBaseClass);													// Add to hash table
		numToRel.addElement(m_relationBaseClass);

		RelationClass rc = new RelationClass(CONTAIN_ID, numRelationClasses++, m_relationBaseClass, this);								// contain
		m_relationClasses.put(CONTAIN_ID, rc);																							// Add to hash table
		numToRel.addElement(rc);

		m_supplierSet   = new SupplierSet(this);
		m_clientSet     = new ClientSet(this);	// Ignore suppliers when computing clients

		addMouseMotionListener(this);
		// The default empty instance

		// No marks are needed when the diagram is first drawn
		setPreserveEntityMarks(0);
		setPreserveRelationMarks(0);
		navigateTo(m_rootInstance);
	}

	protected Vector bgpaths;

	public void run() {

		Object ncontext;

		Enumeration en = bgpaths.elements();

		while (en.hasMoreElements()) {
			String path = (String) en.nextElement();

			if (Util.isHTTP(path)) {
				ncontext = parseURL(path, (URL) context);
			} else {
				ncontext = parseFile(path, (File) context);
			}
		}
	}

	protected String backgroundLoad(Vector paths) {

		if (paths != null && paths.size() > 0) {
			bgpaths = paths;
			Thread thread = new Thread(this);
			thread.start();
		}
		return null;
	}

	// This logic attempts to assign a Rectangle to any component which lacks Rectangle information

	// Called from landscapeEditorCore.attach()
	// Called from landscapeEditorCore.loadLs()
	// Called from landscapeViewerCore.initialLoad()

	// Returns non-null string on error

	public String loadDiagram(String taPath, Vector bgPaths, Object context) {

		Attribute	attr;
		boolean		option;

		resString = null;
		char lc	  = taPath.charAt(taPath.length()-1);


		if (lc == File.separatorChar) {
			taPath = taPath.substring(0, taPath.length()-1);
		}


		// sourceList = new Vector();

		if (taPath.equals("__internal")) {
			processInternalBuffer(taPath);		
		} else if (context instanceof URL || Util.isHTTP(taPath)) {
			MsgOut.dprintln("Parse a URL");
			this.context = parseURL(taPath, (URL) context);
		} else {
			MsgOut.dprintln("Parse a file");
			this.context = parseFile(taPath, (File) context);
		}

		attr = m_rootInstance.getAttribute(EDGEMODE_ID);
		if (attr.avi == null) { 
			attr.avi = new AttributeValueItem(String.valueOf(Do.DIRECT_EDGE_STATE));
		}

		LandscapeObject.setEdgeMode(attr.parseInt());

		attr = m_rootInstance.getAttribute(TOPCLIENTS_ID);
		if (attr.avi == null) {
			option   = Do.TOP_CLIENTS_DEFAULT;
			attr.avi = new AttributeValueItem(String.valueOf(option));
		} else {
			option   = attr.parseBoolean();
		}
		m_ls.setTopClients(option);

		attr = m_rootInstance.getAttribute(WANTCLIENTS_ID);
		if (attr.avi == null) {
			option   = Do.SHOW_CLIENTS_DEFAULT;
			attr.avi = new AttributeValueItem(String.valueOf(option));
		} else {
			option   = attr.parseBoolean();
		}
		m_ls.setShowClients(option);

		attr = m_rootInstance.getAttribute(WANTSUPPLIERS_ID);
		if (attr.avi == null) {
			option   = Do.SHOW_SUPPLIERS_DEFAULT;
			attr.avi = new AttributeValueItem(String.valueOf(option));
		} else {
			option   = attr.parseBoolean();
		}
		m_ls.setShowSuppliers(option);

		attr = m_rootInstance.getAttribute(WANTCARDINALS_ID);
		if (attr.avi == null) {
			option   = Do.SHOW_CARDINALS_DEFAULT;
			attr.avi = new AttributeValueItem(String.valueOf(option));
		} else {
			option   = attr.parseBoolean();
		}
		m_ls.setShowCardinals(option);

		attr = m_rootInstance.getAttribute(RELN_HIDDEN_ID);
		relVisibilityAttr = attr;

		AttributeValueItem avi = attr.avi;

		while(avi != null) {
			RelationClass rc = (RelationClass) m_relationClasses.get(avi.value);
			rc.setClassVisible(false); 
			avi = avi.next; 
		}

		setVisibilityFlags();

		if (this.context == null) {
			return resString;
		}
		return backgroundLoad(bgPaths);
	}

	// Save the landscape in the output stream

	public void saveDiagram(OutputStream os, boolean markEnd) throws IOException
	{
		BufferedOutputStream bos = new BufferedOutputStream(os);
		PrintStream ps = new PrintStream(bos);

		writeSchemeTuples(ps);
		writeSchemeAttributes(ps);
		writeFactTuples(ps);
		writeFactAttributes(ps);

		if (markEnd) {
			ps.print("END\n");
		}
		ps.flush();
		ps.close();
		changedFlag = false;
	}

	/* writeUTF prepends a short to the output specifying the length of the output
     * for characters > 127 multibyte sequence is used
	 *
     * Key:
     *		<l> length of UTF string (always a two byte encoded short)
	 *		<i>	integer value
	 *		<s>	UTF string
	 *
	 * Outputs:  
	 *
	 *		<l>#TA_RAW#						-- header
	 *		<i>								-- size of string table
	 *		<s>*							-- repeats
	 *
	 *      <l>#END#
	 */

	public void	saveDiagramRaw(OutputStream os) throws IOException
	{
		Enumeration en;
		BufferedOutputStream bos = new BufferedOutputStream(os);
		DataOutputStream	 dos = new DataOutputStream(bos);

		Hashtable stringHash = new Hashtable();
		Vector	  stringTable = new Vector();
		String str;

		int id = 0;

		for (en = m_entityClasses.keys(); en.hasMoreElements(); ) {
			str = (String) en.nextElement();
			stringHash.put(str, new Integer(id++));
			stringTable.addElement(str);
		}
		
		for (en = m_relationClasses.keys(); en.hasMoreElements(); ) {
			str = (String) en.nextElement();
			stringHash.put(str, new Integer(id++));
			stringTable.addElement(str);
		}

		for (en = m_entityInstances.keys(); en.hasMoreElements(); ) {
			str = (String) en.nextElement();
			stringHash.put(str, new Integer(id++));
			stringTable.addElement(str);
		}

		for (en = m_attrNames.keys(); en.hasMoreElements(); ) {
			str = (String) en.nextElement();
			stringHash.put(str, new Integer(id++));
			stringTable.addElement(str);
		}

		dos.writeUTF(RAW_HEADER);

		/* Write string table */

		dos.writeInt(stringTable.size());
		for (en = stringTable.elements(); en.hasMoreElements(); ) {
			dos.writeUTF((String) en.nextElement());
		}

		writeSchemeTuplesRaw(dos, stringHash);
		writeSchemeAttributesRaw(dos, stringHash);
		writeFactTuplesRaw(dos, stringHash);
		writeFactAttributesRaw(dos, stringHash);

		dos.writeUTF("#END#");
		dos.flush();
		dos.close();
		bos.close();

		changedFlag = false;
	}

	protected void doSaveForUndo() 
	{
		m_undoDg.m_ls = m_ls;
		m_undoDg.m_entityClasses    = (Hashtable) m_entityClasses.clone();
		m_undoDg.m_relationClasses	= (Hashtable) m_relationClasses.clone();
		undoValid = true; 
	}

	// Make a backup of the current state

	public void saveForUndo() 
	{
		changedFlag = true;
		m_rootInstance.saveForUndo();
		doSaveForUndo();
	}

	public void undo() 
	{
		m_rootInstance.undo();
		Diagram undoDg = m_undoDg;
		m_undoDg = new Diagram(); 
		doSaveForUndo();						// Save the undo inverse 
		m_entityClasses   = m_undoDg.m_entityClasses; 
		m_relationClasses = m_undoDg.m_relationClasses; 
	}

	public boolean isUndoAvailable() 
	{
		return undoValid;
	}

	public Graphics getContext(Graphics gc) 
	{
		int vpwidth, vpheight;

		vpwidth = getWidth()-GAP*2;
		vpheight = getHeight() - GAP*2;
		return gc.create(getX()+GAP, getY()+GAP, vpwidth, vpheight);
	}

	public void rescaleDiagram() 
	{
//		System.out.println("Diagram.rescaleDiagram");
		m_drawRoot.rescaleChildren();
	}

	// 
	// Generates global coordinates for current scale factor
	//

	public void rescaleDiagram(int x, int y, int width, int height) 
	{
//		System.out.println("Diagram.rescaleDiagram {" + x + ", " + y + " " + width + "x" + height + "}");
		m_drawRoot.setBothBounds(0, 0, x, y, width, height);
	}

	public void deleteEdge(RelationInstance ri)
	{
		ri.getSrc().removeSrcRelation(ri);
		ri.getDst().removeDstRelation(ri);
	}

	public EntityInstance intersects(Rectangle lyt)
	{
		return m_drawRoot.intersects(lyt); 
	}

	public EntityInstance containing(Rectangle lyt) 
	{
		return m_drawRoot.containing(lyt);
	}

	public Vector setGroupRegion(Rectangle lyt) 
	{ 
		EntityInstance e = m_drawRoot.getMouseOver((int) lyt.x, (int) lyt.y);

		if (e == null) {
			return null;
		}
		return e.groupRegion(lyt);
	}

	public void setMode(int mode) 
	{
		this.mode = mode; 
	}


	public void setDrawBends(boolean state) 
	{
		MsgOut.dprintln("Draw bends set to: " + state);
		m_drawBends = state;
	}

	public boolean isDrawBends() 
	{
		return m_drawBends;
	}

	public void setEdgeMode(int mode) 
	{
		LandscapeObject.setEdgeMode(mode);
		Attribute attr = m_rootInstance.getAttribute(EDGEMODE_ID);
		attr.avi = new AttributeValueItem(String.valueOf(mode));
	}

	public int getEdgeMode() 
	{
		return LandscapeObject.getEdgeMode();
	}

	public EntityInstance getDrawRoot() 
	{
		return m_drawRoot; 
	}

	public void setDrawRoot(EntityInstance e)
	{
		m_drawRoot = e;
//		e.setBounds(0,0,0,0);
	}

	public EntityInstance getRootInstance() 
	{
		return m_rootInstance;
	}

	public int getTimeStamp() 
	{
		return timeStamp;
	}

	// Is this the BASE CLASS relation or the contains relation

	public boolean excludeReln (RelationClass rc) 
	{
		String id = rc.getId();

		return id.equals(CONTAIN_ID) || id.equals(RelationClass.RELATION_BASE_CLASS_ID);
	}

	public void toggleDstElision(EntityInstance e) 
	{
		Enumeration en;
		RelationClass rc;

		for (en = enumRelations(); en.hasMoreElements(); ) {
			rc = (RelationClass) en.nextElement(); 
			if (!excludeReln(rc) && rc.isClassVisible()) {
				e.toggleDstElision(rc); 
			}
		}
	}

	public void toggleSrcElision(EntityInstance e) 
	{
		Enumeration en;
		RelationClass rc;

		for (en = enumRelations(); en.hasMoreElements(); ) {
			rc = (RelationClass) en.nextElement();
			if (!excludeReln(rc) && rc.isClassVisible()) {
				e.toggleSrcElision(rc);
		}	}
	}

	public void toggleEnteringElision(EntityInstance e) 
	{

		Enumeration en;
		RelationClass rc;

		for (en = enumRelations(); en.hasMoreElements(); ) {
			rc = (RelationClass) en.nextElement(); 
			if (!excludeReln(rc) && rc.isClassVisible()) {
				e.toggleEnteringElision(rc); 
		}	}
	}

	public void toggleExitingElision(EntityInstance e) 
	{
		Enumeration en;
		RelationClass rc;

		for (en = enumRelations(); en.hasMoreElements();) {
			rc = (RelationClass) en.nextElement();
			if (!excludeReln(rc) && rc.isClassVisible()) {
				e.toggleExitingElision(rc);
		}	}
	}

	public void toggleInternalElision(EntityInstance e) 
	{
		Enumeration en;
		RelationClass rc;

		for (en = enumRelations(); en.hasMoreElements(); ) {
			rc = (RelationClass) en.nextElement();
			if (!excludeReln(rc) && rc.isClassVisible())
				e.toggleInternalElision(rc);
		}
	}

	public void toggleContainElision(EntityInstance e) 
	{
		MsgOut.dprintln("Toggle contains elision on " + e.getLabel());
		e.toggleDstElision(CONTAIN_ID);
		e.openStatusUnknown();
	}

	public boolean setContainElision(EntityInstance e, boolean open) 
	{
		if (e.canOpen() && ((e.isOpen() && !open) || (!e.isOpen() && open))) {
			toggleContainElision(e);
			return true;
		}
		return false;
	}

	public void setVisible(boolean value)
	{
		if (m_drawRoot != null) {
			m_drawRoot.setVisible(value);
	}	}
	
	public void setEdgeVisibilityState(RelationClass rc, boolean state) 
	{
		rc.setClassVisible(state);

		// List is HIDDEN relations

		if (state) {
			relVisibilityAttr.removeFromList(rc.getId());
		} else {
			relVisibilityAttr.addToList(rc.getId());
		}
		setVisibilityFlags();
	}

	public void toggleEdgeVisibility() 
	{
		Enumeration en;
		RelationClass rc;

		for (en = enumRelations(); en.hasMoreElements();) {
			rc = (RelationClass) en.nextElement();
			if (!excludeReln(rc) && rc.isClassVisible()) {
				setEdgeVisibilityState(rc, false);
			}
		}
	}

	// Only draw the edges found

	public void clearDrawEdges() 
	{
//		System.out.println("clearDrawEdges");
		m_drawEdges = false;
//		java.lang.Thread.dumpStack();
//		System.out.println("-----");
	}

	// Draw all edges

	public void setDrawEdges() 
	{
//		System.out.println("setDrawEdges");
		m_drawEdges = true;
//		java.lang.Thread.dumpStack();
//		System.out.println("-----");
	}

	public boolean getDrawEdges()
	{
		return(m_drawEdges);
	}

	public Vector getHighlightedEdges() 
	{
		Vector v = new Vector();

		m_drawRoot.getHighlightedEdges(v);
		if (v.size() == 0) {
			return null;
		}
		return v;
	}

	public boolean clearFlags() 
	{
		boolean ret;

		m_keyEntity = null;

		ret = false;
		if (!getDrawEdges()) {
			ret = true;
			setDrawEdges();
		}
		m_rootInstance.clearAllFlags();
		return(ret);
	}

	// Remove all RedBox and HighLight flags under the root

	public void clearQueryFlags() 
	{
		m_rootInstance.clearQueryFlags();
	}

	public void clearGroupFlags() 
	{
		m_keyEntity = null;
		setDrawEdges();
		m_rootInstance.clearGroupFlags();
	}

	public boolean toggleQueryPersistence() 
	{
		persistentQuery = !persistentQuery;
		return persistentQuery;
	}

	public boolean getQueryPersistance() 
	{
		return persistentQuery;
	}

	public boolean edgeSrcValid(EntityInstance e) 
	{
/*
		return rc.isValidSrc(e.getEntityClass()); 
*/
		return true;
	}

	public boolean edgeDstValid(EntityInstance e) 
	{
/*
		return rc.isValidDst(e.getEntityClass()); 
*/
		return true;
	}

	public boolean edgePresent(EntityInstance src, EntityInstance dst) 
	{
/*
		return src.isRelationPresent(rc, dst); 
*/
		return false;
	}

	public void addEdge(EntityInstance src, EntityInstance dst, RelationClass rc)
	{

/*
		String key = Util.hashEdge(rc, src, dst);
		if (edges.contains(key))
			return;
		edges.put(key, key);
*/
		RelationInstance ri = rc.newRelation(src, dst, this);
		src.addRelation(ri, dst); 
	}

	public void newEntity(EntityInstance container)
	{
		EntityInstance	e;
		EntityClass		ec;
		int				n;
		String			ename;

		for (n = 0; ; ++n) {
			ename = "Entity#" + n;
			if (!entityExists(ename)) {
				break;
		}	}

		e = new BaseEntity(defaultEntityClass, ename, this);
		e.setNid(m_entityInstances.size());
		m_entityInstances.put(ename, e);
		ec = e.getEntityClass();
		if (ec != null) {
			ec.makeInstanceOfUs(e);
		}
		e.setxRelLocal(0.25);
		e.setyRelLocal(0.25);
		e.setwidthRelLocal(0.5);
		e.setheightRelLocal(0.5);
		e.rescale(container.getDiagramX(), container.getDiagramY(), container.getWidth(), container.getHeight());

		container.addContainment(e);
		prepostorder();
		EditAttribute.Create(m_ls, e);
	}

	public EntityInstance addEntity(String id, String label, String desc, Rectangle lyt, EntityInstance container,	EntityClass ec, Vector contents)
	{
		Enumeration		en;
		EntityInstance	e, child;
		int				diagramX, diagramY;

		e = ec.newEntity(id, this); 
		e.setNid(m_entityInstances.size());
		m_entityInstances.put(id, e);

		container.addContainment(e);


		e.setDiagramBounds(lyt);
		e.setLabel(label);
		e.setDescription(desc);
		diagramX = e.getDiagramX();
		diagramY = e.getDiagramY();

		for (en = contents.elements(); en.hasMoreElements(); ) {
			child = (EntityInstance) en.nextElement();

			Rectangle rect = child.getDiagramBounds();
			container.removeContainment(child);
			e.addContainment(child);
			child.setBothBounds(diagramX, diagramY, rect.x - e.getDiagramX(), rect.y - e.getDiagramY(), rect.width, rect.height);
		}

		m_rootInstance.rescaleChildren();
		return e;
	} 

	public EntityInstance addEntity(String initId, Rectangle lyt, EntityInstance container)
	{
		EntityInstance	e;
		EntityClass		parent;

		e = new BaseEntity(defaultEntityClass, initId, this);
		e.setDiagramBounds(lyt);
		e.setNid(m_entityInstances.size());
		m_entityInstances.put(initId, e);

		EditAttribute.Create(m_ls, e);
		parent = e.getEntityClass();

		if (parent != null) {
			parent.makeInstanceOfUs(e);
		}
		container.addContainment(e);
		prepostorder();
		m_rootInstance.rescaleChildren();
		return e;
	}

	public RelationClass nameToRelationClass(String name) 
	{
		if (m_relationClasses.containsKey(name)) {

		   return (RelationClass) m_relationClasses.get(name);

		}
		return null;
	}

	public Enumeration enumRelations() 
	{
		return new OrderedHashTableEnumeration(m_relationClasses); 
	}

	public int numRelationClasses() 
	{
		return this.numRelationClasses;
	}

	public int numVisibleRelationClasses() 
	{
		return this.numVisibleRelationClasses;
	}

	public RelationClass getContainsClass() 
	{
		return (RelationClass) m_relationClasses.get(CONTAIN_ID); 
	}



	public Enumeration enumEntityClasses() 
	{
		return new OrderedHashTableEnumeration(m_entityClasses); 
	}



	public int numEntityClasses() 
	{
		return m_numEntityClasses;
	}



	public Enumeration enumEntities() 
	{
		return m_entityInstances.elements();
	}

	public Object getContext() 
	{
		return context;
	}



	public Object setContext(Object context) 
	{
		return this.context = context;
	}



	public String getName() 
	{
		if (context instanceof File) {
			return Util.nameFromPath(((File) context).getPath()); 
		}
		return m_rootInstance.getLabel();
	}



	public String getDir(File file) 
	{
		if (file.isAbsolute()) {
			return file.getParent();
		}
		return (new File(file.getAbsolutePath())).getParent();
	}



	public String getDir() 
	{
		return getDir((File) context);
	}

	public String getAbsolutePath() 
	{
		if (context instanceof File) {
			return ((File) context).getAbsolutePath();
		}
		return ((URL) context).toExternalForm();
	}

	public Vector getGroup() 
	{
		Vector grp = new Vector();

		m_rootInstance.getGroup(grp);
		if (grp.isEmpty()) {
			return null;
		}
		return grp;
	}

	public Rectangle getGroupBoundingBox() 
	{
		Vector grp = getGroup();

		if (grp == null || grp.isEmpty()) {
			return null;
		}

		// Determine the bounding box of whole group

		Enumeration enum1;
		double x1, y1, x2, y2;


		x1 = Double.MAX_VALUE;
		y1 = Double.MAX_VALUE;
		x2 = Double.MIN_VALUE;
		y2 = Double.MIN_VALUE;

		for (enum1 = grp.elements(); enum1.hasMoreElements(); ) {
			EntityInstance e = (EntityInstance) enum1.nextElement();
			Rectangle lyt = e.getDiagramBounds();

			if (lyt.x < x1) 
				x1 = lyt.x;

			if (lyt.y < y1)
				y1 = lyt.y;

			if (lyt.x + lyt.width > x2)
				x2 = lyt.x + lyt.width;

			if (lyt.y + lyt.height > y2)
				y2 = lyt.y + lyt.height;
		}
		return new Rectangle((int) x1, (int) y1, (int) (x2-x1), (int) (y2-y1));
	}



	public EntityInstance getKeyEntity() 
	{
		return m_keyEntity;
	}

	public void setKeyEntity(EntityInstance e) {

		if (m_keyEntity != null) {
			m_keyEntity.clearGroupKeyFlag();		// Must do this first
			m_keyEntity.setGroupFlag();
		}
		m_keyEntity = e;
		m_keyEntity.setGroupKeyFlag();
		m_keyEntity.setGroupFlag();
	}

	public Vector getRedBoxGroup() 
	{
		Vector grp = new Vector();

		m_rootInstance.getRedBoxGroup(grp);
		if (grp.isEmpty()) {
			return null;
		}
		return grp;
	}

	public void selectEdges(Vector edges) 
	{
		Enumeration en;
		RelationInstance r;

		for (en = edges.elements(); en.hasMoreElements(); ) {
			r = (RelationInstance) en.nextElement();
			r.setGroupAndHighlightFlag();
		}
		clearDrawEdges();
	}

	public boolean isOverGroupRegion(int x, int y) 
	{
		Rectangle lyt = getGroupBoundingBox();

		if (lyt != null) {
			return lyt.contains(x, y);
		}
		return false;
	}

	public int getGrid() 
	{
		return gridPixels;
	}

	public void setGrid(int val) 
	{
		gridPixels = val;
	}

	public void flagChanged() {

		changedFlag = true;

	}

	public boolean getChangedFlag()
	{
		return changedFlag;
	}

	public void resetChangedFlag() 
	{
		changedFlag = false;
	}

	public void setToViewport() 
	{
		navigateTo();
	}

	public void fitTo(EntityInstance e) 
	{
		e.fitTo(true);
		rescaleDiagram();
	}

	// Mark all edges as needing to be repainted

	public void invalidateAllEdges(EntityInstance root)
	{
		Enumeration en;

		/* Draw all the visible objects */

		root.invalidateAllEdges(); 

		if (root.isOpen()) {
			for (en = root.getChildren(); en.hasMoreElements(); ) {
				invalidateAllEdges((EntityInstance) en.nextElement());
		}	}
	}

	// Draw all edges marked as needing to be repainted at or under e

	protected void drawHighlightEdges(EntityInstance e)
	{
		Enumeration en;

		/* Draw all the highlight edges */

		e.drawHighlightedEdges();
		if (e.isOpen()) {
			for (en = e.getChildren(); en.hasMoreElements(); ) {
				drawHighlightEdges((EntityInstance) en.nextElement());
		}	}
	}
	// Draw all edges marked as needing to be repainted at or under e

	protected void drawAllEdges(EntityInstance e, EntityInstance under)
	{
		Enumeration en;

		/* Draw all the visible edges */

		e.drawAllEdges(under); 
		if (e.isOpen()) {
			for (en = e.getChildren(); en.hasMoreElements(); ) {
				drawAllEdges((EntityInstance) en.nextElement(), under);
		}	}
	}

	protected void doDrawEdges(EntityInstance e, EntityInstance under)
	{
		/* Draw all the visible objects */

		if (getDrawEdges()) {
			if (m_visibleEdges) {
				drawAllEdges(e, under); 
			}
		} else {
			drawHighlightEdges(e);
	}	}

/* All complex stuff is delayed until we validate to ensure that we have a
   graphic object so that we can compute sizes of labels etc. Much of this
   stuff doesn't really belong in validate -- which is supposed to be called
   only whenever the diagram changes its shape.
 */

	public void validate()
	{
		if (!m_loaded) {
			int				mw, mh, yshift;
			Vector			clients, suppliers;
			Enumeration		en;
			Container		container;
			EntityInstance	e;
			boolean			isTopClients;
			int				temp;
			Insets			insets;

			container = getParent();
			if (container == null) {
				return;
			}
			insets = container.getInsets();

//			System.out.println("Started validation");
			
			mw = container.getWidth()-insets.right-insets.left;
			if (mw <= 0) {
				return;
			}
			mh = container.getHeight()-insets.bottom-insets.top;
			if (mh <= 0) {
				return;
			}

			mw *= m_zoom_x;
			mh *= m_zoom_y;

			m_preferredSize.setSize(mw, mh);
			setPreferredSize(m_preferredSize);

//			System.out.println("Diagram.validate() " + getBounds());
/*			java.lang.Thread.dumpStack();
			System.out.println("-----");
*/
			m_ls.waitCursorOn();

			setBounds(insets.top, insets.left, mw, mh);
			if (m_cardinals == null) {
				m_cardinals = new Container();
				m_cardinals.setLayout(null);
			}
			if (m_edges == null) {
				m_edges = new Container();
				m_edges.setLayout(null);
			}

			removeAll();
			m_cardinals.removeAll();

			// These go on top
			if (getLs().isShowCardinals()) {
				m_cardinals.setBounds(insets.top, insets.left, mw, mh);
				add(m_cardinals);
			}
			// This goes under cardinals if present
			m_edges.removeAll();
			m_edges.setBounds(insets.top, insets.left, mw, mh);
			add(m_edges);		// Add the set of all edges


			// Clear all but the indicated marks

			if (!getDrawEdges()) {
				m_preserve_entity_marks   |= EntityInstance.REDBOX_MARK;
				m_preserve_relation_marks |= RelationInstance.HIGHLIGHT_FLAG_MARK;
			}

			m_rootInstance.clearAllMarks(m_preserve_entity_marks, m_preserve_relation_marks);

//			System.out.println("Diagram.validate() cleared all marks");

			/* This has to be delayed until we have resized things appropriately since the adding of
			   the drawRoot to the diagram causes the edges to also be added, and these need to know
			   displayable dimensions of entities
			 */

			mw       -= GAP*6;		// Maximum width  of e
			mh	     -= GAP*4;		// Maximum height of e
//			System.out.println("mw=" + mw + " mh=" + mh);
			yshift    = 0;
			clients   = null;
			suppliers = null;
			isTopClients = m_ls.isTopClients();

			// Make entity take a full footprint in the display.

			if (m_drawRoot != m_rootInstance) {
				add(m_exitFlag);
				m_exitFlag.setVisible(true);
			} else {
				// If we don't do a repaint the [-] box doesn't vanish
				repaint();
			}

//			System.out.println("Diagram.validate() added to Diagram");

			/* This performs a first cut at finding clients and suppliers
			 * It essentially collects into a vector topmost clients and suppliers
			 */

			findClientsAndSuppliers();

//			System.out.println("Diagram.validate() found clients and suppliers");

			if (m_clientSet.getFullSetSize() != 0) {
				// we will either show or report not showing
				// Reduce height to allow clients
				temp = (ClientSupplierSet.CLIENT_SUPPLIER_HEIGHT + GAP*4);
				mh  -= temp;
				if (isTopClients) {
					m_clientSet.setBounds(GAP*3, GAP*2, mw, ClientSupplierSet.CLIENT_SUPPLIER_HEIGHT);
					yshift = temp; 
				} else {
					m_clientSet.setBounds(GAP*3, getHeight() - ClientSupplierSet.CLIENT_SUPPLIER_HEIGHT - GAP*2, mw, ClientSupplierSet.CLIENT_SUPPLIER_HEIGHT);
				}
//				System.out.println("Diagram.validate() m_clientSet=" + m_clientSet.getBounds());
			}

			if (m_supplierSet.getFullSetSize() != 0) {			// Reduce height to allow suppliers
				temp = (ClientSupplierSet.CLIENT_SUPPLIER_HEIGHT + GAP*4);
				mh  -= temp;
				if (!isTopClients) {
					m_supplierSet.setBounds(GAP*3, GAP*2, mw, ClientSupplierSet.CLIENT_SUPPLIER_HEIGHT);
					yshift = temp; 
				} else {
					m_supplierSet.setBounds(GAP*3, getHeight() - ClientSupplierSet.CLIENT_SUPPLIER_HEIGHT - GAP*2, mw, ClientSupplierSet.CLIENT_SUPPLIER_HEIGHT);

//					System.out.println("Supplier=" + (getY() + m_supplierSet.getY() + m_supplierSet.getHeight()) + " actual=" + m_ls.getHeight());
				}
//				System.out.println("Diagram.validate() m_supplierSet=" + m_supplierSet.getBounds());
			}

			ClientSupplierSet.compact(m_clientSet, m_supplierSet);

			clients = m_clientSet.getFullSet();
			if (clients != null && clients.size() == 0) {
				clients = null;
			}

			suppliers = m_supplierSet.getFullSet();
			if (suppliers != null && suppliers.size() == 0) {
				suppliers = null;
			}
			rescaleDiagram(GAP*3, GAP*2+yshift, mw, mh);
//			System.out.println("Diagram.validate() rescaled Diagram");

			m_drawRoot.addToDiagram(this);

			for (en = m_drawRoot.getChildren(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.computeAllDiagramEdges(m_drawRoot, clients, suppliers, null /* No closed entity yet */);
			}
//			System.out.println("Diagram.validate() computed all diagram edges");


			for (en = m_drawRoot.getChildren(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.liftAllDiagramEdges(m_drawRoot, clients, suppliers, null /* No closed entity yet */);
			}
//			System.out.println("Diagram.validate() lifted all edges");

			// Do after lifted edges since ordering based on where edges go to/come from

			if (m_supplierSet.getFullSetSize() != 0) {	
				m_supplierSet.order();
				add(m_supplierSet);		// Add the quasi supplier box to diagram
				m_supplierSet.setVisible(true);
//				System.out.println("Diagram.validate() added Suppliers");
			}

			if (m_clientSet.getFullSetSize() != 0) {
				m_clientSet.order();
				add(m_clientSet);		// Add the quasi client box to the diagram
				m_clientSet.setVisible(true);
//				System.out.println("Diagram.validate() added Clients");
			}

			doDrawEdges(m_drawRoot, m_drawRoot);	// Marks have been previously cleared so do not need to invalidate

			if (getLs().isShowCardinals()) {
				BaseEntity	drawInstance;
			
				drawInstance = (BaseEntity) m_drawRoot;
				drawInstance.resetCardinals(numRelationClasses);
				drawInstance.calcEdgeCardinals(null, !getDrawEdges());
				drawInstance.showCardinals();
//				System.out.println("Show cardinals " + m_cardinals.getComponentCount());
			}

//			System.out.println("Diagram.validate() processed edges");

			MapBox mapBox;


			mapBox = m_ls.m_mapBox;
			if (mapBox != null) {
				mapBox.validate();
				mapBox.setVisible(true);
//				System.out.println("Diagram.validate() map box added");

			}

			m_ls.waitCursorOff();

			m_loaded = true;
			repaint();

//			System.out.println("Diagram.validate() done");
		}	
	}

	public void redrawDiagram()
	{
		setLoaded(false);
		validate();
	}

	public boolean zoom(double xfactor, double yfactor) 
	{
		if (xfactor <= 1.0 && yfactor <= 1.0 && m_zoom_x == 1.0 && m_zoom_y == 1.0) {
			return(false);
		}
		m_zoom_x = m_zoom_x * xfactor;
		m_zoom_y = m_zoom_y * yfactor;

		if (m_zoom_x < 1.0) {
			m_zoom_x = 1.0;
		}
		if (m_zoom_y < 1.0) {
			m_zoom_y = 1.0;
		}

		redrawDiagram();
		revalidate();
		return true;
	}

	public boolean set_to_viewport()
	{
		if (m_zoom_x == 1.0 || m_zoom_y == 1.0) {
			return(false);
		}
		m_zoom_x = 1.0;
		m_zoom_y = 1.0;
		redrawDiagram();
		revalidate();
		return true;
	}

	// Carefully decide on values for preserveEntityMarks and preserveRelationMarks before calling this

	public void navigateTo(EntityInstance e) 
	{
		// Make entity take a full footprint in the display.

		setDrawRoot(e);
		redrawDiagram();
	}

	public void navigateTo(String eid) 
	{
		EntityInstance e = (EntityInstance) m_entityInstances.get(eid);

		if (e != null) {
			navigateTo(e);
	}	}

	// Navigate to default entity

	public void navigateTo()
	{
		Enumeration		en;
		EntityInstance	e;
			
		en = m_rootInstance.getChildren();
		e  = m_rootInstance; 
		if (en.hasMoreElements()) {
			e = (EntityInstance) en.nextElement();
			if (en.hasMoreElements()) {
				e = m_rootInstance;
		}	}

		setPreserveEntityMarks(0);
		setPreserveRelationMarks(0);
		navigateTo(e);
	}

	public void navigateToDrawRootParent()
	{
		EntityInstance e;

		e = m_drawRoot.getEnterableParent();

		if (e == null) {
			m_ls.error("At topmost landscape");
		} else {
			// Going up
			setPreserveEntityMarks(0);
			setPreserveRelationMarks(0);
			navigateTo(e);
	}	}

	public boolean isReadOnly() 
	{
		return readOnly;
	}

	public void setReadOnly()
	{
		readOnly = true;
	}



	public boolean entityExists(String name) 
	{
		EntityInstance e = (EntityInstance) m_entityInstances.get(name);

		return (e != null);
	}

	public LandscapeViewerCore getLs() 
	{
		return m_ls;
	}

	public EntityInstance getClient(EntityInstance e) 
	{ 
		EntityInstance	pe;

		for (pe = e; pe != null; pe = pe.getContainedBy()) {
			if (pe.isClient()) {
				return(pe);
		}	}
		return(null);
	}



	public EntityInstance getSupplier(EntityInstance e) 
	{	
		EntityInstance pe;


		for (pe = e; pe != null; pe = pe.getContainedBy()) {
			if (pe.isSupplier()) {
				return(pe);
		}	}
		return(null);
	}

	public EntityInstance getClientOrSupplier(EntityInstance e) 
	{
		EntityInstance pe;


		for (pe = e; pe != null; pe = pe.getContainedBy()) {
			if (pe.isClientOrSupplier()) {
				return(pe);
		}	}
		return(null);
	}

	public boolean isClientDescendent(EntityInstance e) 
	{
		return (getClient(e) != null);
	}



	public boolean isSupplierDescendent(EntityInstance e) 
	{
		return (getSupplier(e) != null);
	}

	public RelationClass numToRelationClass(int n) 
	{
		return (RelationClass) numToRel.elementAt(n);
	}

	public boolean allowElision() 
	{
		return getDrawEdges();
	}

	protected boolean m_modeHandlingActive = false;

	protected long lastUp = 0;

	protected LandscapeModeHandler m_modeHandler;
	
	public void setModeHandler(LandscapeModeHandler modeHandler)
	{
//		System.out.println("Diagram.setModeHandler " + modeHandler);
		if (modeHandler != m_modeHandler) {
			if (modeHandler != null) {
				// Make sure mode handler knows it is talking to us before
				// we use it.
				modeHandler.select(this);
			}
			m_modeHandler = modeHandler;
	}	}

	public LandscapeModeHandler getModeHandler()
	{
		return(m_modeHandler);
	}

	public boolean isModeHandlingActive()
	{
		return(m_modeHandlingActive);
	}

	public void setModeHandlingActive(boolean value)
	{
		m_modeHandlingActive = value;
	}

	public boolean clipboardDump = false;

	public RelationInstance targetRelation(Object object)
	{
		if (object != null) {
			if (object instanceof Vector) {
				Vector v = (Vector) object;
				switch (v.size()) {
				case 0:
					m_ls.doFeedback("Target relation set is empty");
					break;
				case 1:
					return((RelationInstance) v.firstElement());
				default:
					m_ls.doFeedback("Target relation set has multiple objects in it");
				}
				return(null);
			}
			return((RelationInstance) object);
		}

		Vector v = getHighlightedEdges();

		if (v == null) {
			m_ls.doFeedback("Highlight the edge you wish to work on");
		} else {
			if (v.size() == 1) {
				return((RelationInstance) v.firstElement());
			}
			m_ls.doFeedback("Can't operate on " + v.size() + " highlight edges");
		}
		return(null);
	}

	public EntityInstance targetEntity(Object object)
	{
		if (object != null) {
			if (object instanceof Vector) {
				Vector v = (Vector) object;
				switch (v.size()) {
				case 0:
					m_ls.doFeedback("Target entity set is empty");
					break;
				case 1:
					return((EntityInstance) v.firstElement());
				default:
					m_ls.doFeedback("Target entity set has multiple objects in it");
				}
				return(null);
			}
			return((EntityInstance) object);
		}

		EntityInstance ke;

		ke = getKeyEntity();
		if (ke != null) {
			return(ke);
		}
		return(getDrawRoot());
//		m_ls.doFeedback("Highlight the entity you wish to work on");
//		return(null);
	}

	public Vector targetRelations(Object object)
	{
		Vector v;

		if (object != null) {
			if (object instanceof Vector) {
				v = (Vector) object;
				if (v.size() > 0) {
					return(v);
				}
				m_ls.doFeedback("Target relation set is empty");
				return(null);
			}
			v = new Vector();
			v.addElement(object);
			return(v);
		}

		v = getHighlightedEdges();

		if (v == null) {
			m_ls.doFeedback("Select the edges you wish to work on");
		}
		return(v);
	}

	public Vector targetEntities(Object object)
	{
		Vector v;

		if (object != null) {
			if (object instanceof Vector) {
				v = (Vector) object;
				if (v.size() > 0) {
					return(v);
				}
				m_ls.doFeedback("Target entity set is empty");
				return(null);
			}
			v = new Vector();
			v.addElement(object);
			return(v);
		}
		return(getGroup());
	}

	public void entityPressed(MouseEvent ev, EntityInstance e, int x, int y)
	{
		if (clipboardDump) {
			((LandscapeEditorCore) m_ls).dumpClipboard(e, x, y);
			clipboardDump = false;
			m_ls.setCursor(Cursor.DEFAULT_CURSOR);
			if (m_modeHandler != null) {
				m_modeHandler.movedOverThing(ev, e, x, y);
			}
		} else {
			m_ls.showDescription(e, true);
			if (m_modeHandler != null) {
				setModeHandlingActive(true);
				if (!m_modeHandler.entityPressed(ev, e, x, y)) {
					m_modeHandler.cleanup();
					setModeHandlingActive(false);
	}	}	}	}

	public void entityDragged(MouseEvent ev, EntityInstance e, int x, int y)
	{		
		if (isModeHandlingActive()) {
			if (m_modeHandler.entityDragged(ev, e, x, y)) {
				return;
		}	}
		movedOverThing(ev, e, x, y); // Must remain as unadjusted values!
	}

	public void entityReleased(MouseEvent ev, EntityInstance e, int x, int y)
	{
//		System.out.println("Diagram.entityReleased=" + e);

		if (isModeHandlingActive()) {
			LandscapeModeHandler modeHandler = getModeHandler();

			modeHandler.entityReleased(ev, e, x, y);
			modeHandler.cleanup();
			setModeHandlingActive(false);
		}

		if (y > 0) {
			getLs().toolButton[0].requestFocus();
			getLs().requestFocus();
	}	}		

	public void relationPressed(MouseEvent ev, RelationInstance ri, int x, int y)
	{
//		System.out.println("Diagram.relationPressed=" + ri);

		m_ls.showDescription(ri, true);
		if (m_modeHandler != null) {
			setModeHandlingActive(true);
			if (!m_modeHandler.relationPressed(ev, ri, x, y)) {
				m_modeHandler.cleanup();
				setModeHandlingActive(false);
	}	}	}
	
	public void relationDragged(MouseEvent ev, RelationInstance ri, int x, int y)
	{		
		if (isModeHandlingActive()) {
			if (m_modeHandler.relationDragged(ev, ri, x, y)) {
				return;
		}	}
		movedOverThing(ev, ri, x, y); // Must remain as unadjusted values!
	}
		
	public void relationReleased(MouseEvent ev, RelationInstance ri, int x, int y)
	{
//		System.out.println("Diagram.relationReleased=" + ri);

		if (isModeHandlingActive()) {
			LandscapeModeHandler modeHandler = getModeHandler();

			modeHandler.relationReleased(ev, ri, x, y);
			modeHandler.cleanup();
			setModeHandlingActive(false);
		}

		if (y > 0) {
			getLs().toolButton[0].requestFocus();
			getLs().requestFocus();
	}	}

	public void movedOverThing(MouseEvent ev, Object thing, int x, int y)
	{
		if (!clipboardDump) {
			m_ls.resetAnticipateCursor();
			if (m_modeHandler != null) {
				m_modeHandler.movedOverThing(ev, thing, x, y);
			}
			m_ls.useAnticipateCursor();
		}
		return;
	}

	// MouseMotionListener interface

	public void mouseDragged(MouseEvent ev)
	{
	}

	public void mouseMoved(MouseEvent ev)
	{
//		System.out.println("Diagram.mouseMoved");
		if (!clipboardDump) {
			movedOverThing(ev, this, ev.getX(), ev.getY());
		} else {
			m_ls.setCursor(Cursor.DEFAULT_CURSOR);
		}
		return;
	}

	// DiagramCoordinates interface

	public int getDiagramX()
	{
		return(0);
	}

	public int getDiagramY()
	{
		return(0);
	}
}

